#ifndef __AIRINV_AIRINV_TYPES_HPP
#define __AIRINV_AIRINV_TYPES_HPP

// //////////////////////////////////////////////////////////////////////
// Import section
// //////////////////////////////////////////////////////////////////////
// STL
#include <map>
// Boost
#include <boost/shared_ptr.hpp>
// StdAir
#include <stdair/stdair_exceptions.hpp>
#include <stdair/stdair_inventory_types.hpp>

namespace AIRINV {

  // Forward declarations
  class AIRINV_Service;
  class AIRINV_Master_Service;


  // ///////// Exceptions ///////////

  /**
   * The inventory input file can not be parsed.
   */
  class InventoryFileParsingFailedException
    : public stdair::ParsingFileFailedException {
  public:
    /**
     * Constructor.
     */
    InventoryFileParsingFailedException (const std::string& iWhat)
      : stdair::ParsingFileFailedException (iWhat) {}
  };

  /**
   * The schedule input file can not be parsed.
   */
  class ScheduleFileParsingFailedException
    : public stdair::ParsingFileFailedException {
  public:
    /**
     * Constructor.
     */
    ScheduleFileParsingFailedException (const std::string& iWhat)
      : stdair::ParsingFileFailedException (iWhat) {}
  };
  
  /**
   * Specific exception when some BOM objects can not be found within
   * the inventory.
   */
  class SegmentDateNotFoundException : public stdair::ParserException {
  public:
    /**
     * Constructor.
     */
    SegmentDateNotFoundException (const std::string& iWhat)
      : stdair::ParserException (iWhat) {}
  };

  /**
   * The inventory input file can not be found or opened.
   */
  class InventoryInputFileNotFoundException : public stdair::FileNotFoundException {
  public:
    /**
     * Constructor.
     */
    InventoryInputFileNotFoundException (const std::string& iWhat)
      : stdair::FileNotFoundException (iWhat) {}
  };

  /**
   * The schedule input file can not be found or opened.
   */
  class ScheduleInputFileNotFoundException : public stdair::FileNotFoundException {
  public:
    /**
     * Constructor.
     */
    ScheduleInputFileNotFoundException (const std::string& iWhat)
      : stdair::FileNotFoundException (iWhat) {}
  };

  /**
   * Duplicated flight date object.
   */
  class FlightDateDuplicationException : public stdair::ObjectCreationgDuplicationException {
  public:
    /**
     * Constructor.
     */
    FlightDateDuplicationException (const std::string& iWhat) 
      : stdair::ObjectCreationgDuplicationException (iWhat) {} 
  };

  /**
   * Specific exception related to bookings made against the inventory.
   */
  class BookingException : public stdair::RootException {
  };


  // //////// Type definitions /////////
  /**
   * (Smart) Pointer on the AirInv (slave) service handler.
   */
  typedef boost::shared_ptr<AIRINV_Service> AIRINV_ServicePtr_T;
  
  /**
   * (Smart) Pointer on the AirInv master service handler.
   */
  typedef boost::shared_ptr<AIRINV_Master_Service> AIRINV_Master_ServicePtr_T;
  
  /**
   * Type defining a map of airline codes and the corresponding
   * airline inventories.
   */
  typedef std::map<const stdair::AirlineCode_T,
                   AIRINV_ServicePtr_T> AIRINV_ServicePtr_Map_T;

  /**
   * Define the FRAT5 curve.
   */
  typedef std::map<const stdair::DTD_T, double> FRAT5Curve_T;

}
#endif // __AIRINV_AIRINV_TYPES_HPP

